/* parse.c - IRC protocol parsing
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: parse.c,v 1.113.2.32 2002/02/22 03:03:04 entrope Exp $
 */

#include "config.h"
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif

#include "common.h"
#include "commands.h"
#include "dict.h"
#include "conf.h"
#include "gline.h"
#include "hash.h"
#include "log.h"
#include "parse.h"
#include "timeq.h"

extern int burst, ping_freq, ping_timeout;
extern struct cManagerNode cManager;
struct userList dead_users;

#define CMD_FUNC(NAME) int NAME(unsigned int argc, unsigned char *argv[])
typedef CMD_FUNC(cmd_func_t);

/* Ping state is kept in the timeq (only one of these two can be in
 * the queue at any given time). */
void
timed_send_ping(void *data)
{
    static void timed_ping_timeout(void *data);
    (void)data;

    irc_ping(self->name);
    timeq_add(now + ping_timeout, timed_ping_timeout, 0);
}

static void
timed_ping_timeout(void *data)
{
    (void)data;

    /* Uplink "health" tracking could be accomplished by counting the
       number of ping timeouts that happen for each uplink. After the
       timeouts per time period exceeds some amount, the uplink could
       be marked as unavalable.*/
    irc_squit(self, "Ping timeout.");
}

static CMD_FUNC(cmd_pass)
{
    const char *true_pass;

    if (argc < 2) return 0;

    true_pass = cManager.uplink->their_password;
    if (true_pass && strcmp(true_pass, argv[1])) {
	/* It might be good to mark the uplink as unavailable when
	   this happens, though there should be a way of resetting
	   the flag. */
	irc_squit(self, "Incorrect password received.");
	return 1;
    }

    cManager.uplink->state = BURSTING;
    send_burst();
    return 1;
}

static CMD_FUNC(cmd_dummy)
{
    (void)argc; (void)argv;
    /* we don't care about these messages */
    return 1;
}

static CMD_FUNC(cmd_error)
{
    (void)argc;

    if (argv[1])
    {
        log(MAIN_LOG, LOG_ERROR, "Error: %s\n", argv[1]);
    }
    log(MAIN_LOG, LOG_ERROR, "Error received from uplink, squitting.\n");

    if(cManager.uplink->state != CONNECTED)
    {
	/* Error messages while connected should be fine. */
	cManager.uplink->flags |= UPLINK_UNAVAILABLE;
	log(MAIN_LOG, LOG_ERROR, "Disabling uplink.\n");
    }

    close_socket();
    return 1;
}

static CMD_FUNC(cmd_server)
{
    if (!argv[0][1] || !argv[0][2]) {
	/* another server introduced us */
	if (argc < 9) return 0;
	AddServer(GetServer(argv[0]), argv[2], atoi(argv[3]), atoi(argv[4]), atoi(argv[5]), argv[7], argv[argc-1]);
    } else {
	/* this must be our uplink */
	if (argc < 8) return 0;
	self->uplink = AddServer(self, argv[1], atoi(argv[2]), atoi(argv[3]), atoi(argv[4]), argv[6], argv[argc-1]);
    }
    return 1;
}

static CMD_FUNC(cmd_stats)
{
    struct userNode *un = GetUser(argv[0]);

    (void)argc;
    if (argc < 4) return 0;

    if (un && !strcmp(argv[2], "u")) irc_uptime(un);
    return 1;
}

static CMD_FUNC(cmd_whois)
{
    struct userNode *from;
    struct userNode *who;

    if (argc < 4) return 0;
    if (!(from = GetUser(argv[0]))) {
        log(MAIN_LOG, LOG_ERROR, "Could not find user for %s\n", argv[0]);
        return 0;
    }
    if (!(who = GetUserH(argv[3]))) {
        return 1;
    }
    irc_whois(from, who);
    return 1;
}

static CMD_FUNC(cmd_eob)
{
    struct server *sender;
    (void)argv; /* we don't care about argv */
    if (argc < 2) return 0;
    sender = GetServer(argv[0]);
    if (burst && sender && sender == self->uplink) {
	burst = 0;
	cManager.uplink->state = CONNECTED;
	irc_eob_ack();
    }
    return 1;
}

static CMD_FUNC(cmd_eob_ack)
{
    extern unsigned long burst_length;
    extern time_t burst_begin;
    (void)argc;(void)argv; /* we don't care about these */

    if (!burst_length) {
        burst_length = now - burst_begin;
    }
    burst = 0;
    cManager.uplink->state = CONNECTED;
    return 1;
}

static CMD_FUNC(cmd_ping)
{
    void received_ping(void);

    if (argc < 3) return 0;
    irc_pong(argv[2]);

    timeq_del(0, timed_send_ping, 0, TIMEQ_IGNORE_WHEN|TIMEQ_IGNORE_DATA);
    timeq_del(0, timed_ping_timeout, 0, TIMEQ_IGNORE_WHEN|TIMEQ_IGNORE_DATA);
    timeq_add(now + ping_freq, timed_send_ping, 0);

    received_ping();
    return 1;
}

struct create_desc {
    struct userNode *user;
    time_t when;
};

static void
create_helper(char *name, void *data)
{
    struct create_desc *cd = data;
    struct chanNode *cn;

    /* We can't assume the channel create was allowed because of the
     * bad-word channel checking.
     */
    cn = AddChannel(name, cd->when, "", "", 0, cd->user->numeric, NULL);

    /* We can't assume the user got ops, or that the single
     * user in the channel is the user who joined either.
     */
    if (cn->members.used == 1) {
	struct modeNode *mn = GetUserMode(cn, cd->user);
	if (mn) mn->modes = MODE_CHANOP;
    }
}

static CMD_FUNC(cmd_create)
{
    struct create_desc cd;
    if (argc < 4) return 0;
    cd.user = GetUser(argv[0]);
    cd.when = atoi(argv[3]);
    parse_foreach(argv[2], NULL, create_helper, NULL, NULL, &cd);
    return 1;
}

static CMD_FUNC(cmd_pong)
{
    void received_ping(void);
    if (argc < 4) return 0;
    if (!strcmp(argv[3], self->name)) {
	timeq_del(0, timed_send_ping, 0, TIMEQ_IGNORE_WHEN|TIMEQ_IGNORE_DATA);
	timeq_del(0, timed_ping_timeout, 0, TIMEQ_IGNORE_WHEN|TIMEQ_IGNORE_DATA);
	timeq_add(now + ping_freq, timed_send_ping, 0);

	received_ping();
    }
    return 1;
}

static CMD_FUNC(cmd_nick)
{
    if (argv[0][1] && argv[0][2]) {
	/* nick change (since the source is a user numeric) */
	if (argc < 3) return 0;
	NickChange(GetUser(argv[0]), argv[2]);
    } else {
	/* new nick */
	if (argc < 10) {
            log(DEBUG_LOG, LOG_ERROR, "expected 10+ args, only got %d\n", argc);
            return 0;
        }
	if (argv[7][0] == '+') {
	    if (argc < 11) return 0;
	    if(!AddUser(GetServer(argv[0]), argv[2], argv[5], argv[6], argv[7], argv[9], argv[10], (time_t)atoi(argv[4]), argv[8]))
	    {
		log(DEBUG_LOG, LOG_ERROR, "could not add user: %s\n", argv[2]);
		return 0;
	    }
	} else {
	    if(!AddUser(GetServer(argv[0]), argv[2], argv[5], argv[6], "+", argv[8], argv[9], (time_t)atoi(argv[4]), argv[7]))
	    {
		log(DEBUG_LOG, LOG_ERROR, "could not add user: %s\n", argv[2]);
		return 0;
	    }
	}
    }
    return 1;
}

static CMD_FUNC(cmd_burst)
{
    char *members = "", *banlist = NULL;
    const char *key = NULL, *modes = NULL;
    unsigned int limit = 0, next = 4;

    if (argc < 4) return 0;
    while (next < argc) {
	switch (argv[next][0]) {
	case '+': {
	    const char *pos;
	    int keypos, limitpos;
	    modes = argv[next++];
	    pos = strchr(modes, 'k');
	    keypos = pos ? (pos-modes) : -1;
	    pos = strchr(modes, 'l');
	    limitpos = pos ? (pos-modes) : -1;
	    if (keypos != -1) {
		if (limitpos != -1) {
		    if (keypos < limitpos) {
			key = argv[next++];
			limit = atoi(argv[next++]);
		    } else {
			limit = atoi(argv[next++]);
			key = argv[next++];
		    }
		} else {
		    key = argv[next++];
		}
	    } else if (limitpos != -1) {
		limit = atoi(argv[next++]);
	    }
	}
	break;
	case '%': banlist = argv[next++]+1; break;
	default: members = argv[next++]; break;
	}
    }

    AddChannel(argv[2], (time_t)atoi(argv[3]), modes, key, limit, members, banlist);

    return 1;
}

static mode_change_func_t *mcf_list;
static unsigned int mcf_size = 0, mcf_used = 0;

void
reg_mode_change_func(mode_change_func_t handler)
{
    if (mcf_used == mcf_size) {
	if (mcf_size) {
	    mcf_size <<= 1;
	    mcf_list = realloc(mcf_list, mcf_size*sizeof(mode_change_func_t));
	} else {
	    mcf_size = 8;
	    mcf_list = malloc(mcf_size*sizeof(mode_change_func_t));
	}
    }
    mcf_list[mcf_used++] = handler;
}

static void
call_mode_change_funcs(struct chanNode *channel, struct userNode *user, int change_type, void *change)
{
    unsigned int n;
    /* someone replace me with a robot. sigh. */
    for (n=0; n<mcf_used; n++)
    {
	mcf_list[n](channel, user, change_type, change);
    }
}

static CMD_FUNC(cmd_mode)
{
    if (argc < 4) return 0;
    if (!IsChannelName(argv[2])) {
	struct userNode *un = GetUserH(argv[2]);
	if (!un) {
            log(MAIN_LOG, LOG_ERROR, "Unable to find user %s whose mode is changing\n", argv[2]);
            return 0;
        }
	mod_usermode(un, argv[3]);
	return 1;
    } else {
	struct chanNode *cn;
	struct userNode *un;
	int next_arg=4, res=1, add=1, bit;
	const char *mchange;
	struct modeNode *mn;
	chan_mode_t mode;

        cn = GetChannel(argv[2]);
        /* If we get a server numeric, source should be NULL. */
        un = argv[0][1] ? GetUser(argv[0]) : NULL;
	if (!cn) {
            log(MAIN_LOG, LOG_ERROR, "Unable to find channel %s whose mode is changing\n", argv[2]);
            return 0;
        }
	mchange = argv[3];
	mode = cn->modes;
	while (*mchange) {
	    switch (*mchange++) {
	    case 0: break;
	    case '+': add = 1; break;
	    case '-': add = 0; break;
#define do_chan_mode(FLAG) do { if (add) mode |= FLAG; else mode &= ~FLAG; } while(0)
	    case 'n': do_chan_mode(MODE_NOPRIVMSGS); break;
	    case 't': do_chan_mode(MODE_TOPICLIMIT); break;
	    case 's': do_chan_mode(MODE_SECRET); break;
	    case 'p': do_chan_mode(MODE_PRIVATE); break;
	    case 'i': do_chan_mode(MODE_INVITEONLY); break;
	    case 'm': do_chan_mode(MODE_MODERATED); break;
            case 'D': do_chan_mode(MODE_DELAYJOINS); break;
#undef do_chan_mode
	    case 'l':
		if (add) {
		    mode |= MODE_LIMIT;
		    cn->limit = atoi(argv[next_arg++]);
		} else {
		    /* We shouldn't need to gobble an argument here
		       (verify this). */
		    mode &= ~MODE_LIMIT;
		}
		break;
	    case 'k':
		if (add) {
		    mode |= MODE_KEY;
		    safestrncpy(cn->key, argv[next_arg++], sizeof(cn->key));
		} else {
		    /* Gobble the key up so following arguments line up. */
		    next_arg++;
		    mode &= ~MODE_KEY;
		}
		break;
	    case 'o':
	    case 'v': {
		struct userNode *vic;
		vic = GetUserN(argv[next_arg++]);
		if (!vic) {
                    log(MAIN_LOG, LOG_ERROR, "Unable to find user with numeric %s\n", argv[next_arg-1]);
                    res = 0;
                    break;
                }
		mn = GetUserMode(cn, vic);
		if (!mn) {
                    log(MAIN_LOG, LOG_ERROR, "Unable to find user %s in channel %s\n", vic->nick, cn->name);
                    res = 0;
                    break;
                }
		bit = (mchange[-1] == 'o') ? MODE_CHANOP : MODE_VOICE;
		if (add) mn->modes |= bit;
		else mn->modes &= ~bit;
		if(bit != MODE_CHANOP) break;

		if(un) call_mode_change_funcs(cn, un, add ? MODE_CHANGE_OP : MODE_CHANGE_DEOP, vic); 
	        break;}
	    case 'b': {
		char *ban = argv[next_arg++];
		if (add)
		{
		    AddChannelBan(1, &ban, cn, argv[0], now, 0);
		    if(un) call_mode_change_funcs(cn, un, MODE_CHANGE_BAN, ban);
		}
		else DelChannelBan(1, &ban, cn, argv[0], 0);
		break;
            }
	    }
	}

	if(mode != cn->modes && un)
	{
	    cn->modes = mode;
	    call_mode_change_funcs(cn, un, MODE_CHANGE_MISC, NULL);
	}
	else
	{
	    cn->modes = mode;
	}
	return res;
    }
}

static CMD_FUNC(cmd_topic)
{
    static struct chanNode *cn;
    static struct userNode *un;
    const char *source;

    if (argc < 4) return 0;
    if (argv[2][0] == '&') return 0;
    cn = GetChannel(argv[2]);
    if (!cn) {
        log(MAIN_LOG, LOG_ERROR, "Unable to find channel %s whose topic is being set\n", argv[2]);
        return 0;
    }

    if (argv[0][0]==':') {
	source = argv[0] + 1;
    } else {
	un = GetUser(argv[0]);
	source = un ? un->nick : (const char*)argv[0];
    }

    SetChannelTopic(cn, source, argv[3], 0);
    return 1;
}

static CMD_FUNC(cmd_num_topic) {
    static struct chanNode *cn;

    if (!argv[1]) return 0;
    if (argv[3]) {
	cn = GetChannel(argv[3]);
	if (!cn) {
            log(MAIN_LOG, LOG_ERROR, "Unable to find channel %s in topic reply\n", argv[3]);
            return 0;
        }
    } else return 0;

    switch (atoi(argv[1])) {
	case 331:
	    cn->topic_time = 0;
	    break;  /* no topic */
        case 332:
	    if (argc < 5) return 0;
	    unsplit_string(argv+4, argc-4);
    	    safestrncpy(cn->topic, argv[4], sizeof(cn->topic));
	    break;
        case 333:
	    if (argc < 6) return 0;
	    /* 332 always precedes 333, so cn->topic is already set */
	    SetChannelTopic(cn, argv[4], NULL, 0);
	    cn->topic_time = atoi(argv[5]); /* SetChannelTopic() uses now */
	    break;
	default:
	    return 0; /* should never happen */
    }
    return 1;
}

static CMD_FUNC(cmd_num_gline)
{
    if (argc < 7) return 0;
    gline_add(argv[0]+1, argv[4], atoi(argv[5])-now, argv[6], 0);
    return 1;
}

static CMD_FUNC(cmd_quit)
{
    struct userNode *user;
    if (argc < 3) return 0;
    user = GetUser(argv[0]);
    /* Sometimes we get a KILL then a QUIT or the like, so we don't want to
     * call DelUser unless we have the user in our grasp.
     */
    if (user) {
        DelUser(user, NULL, false, argv[2]);
    }
    return 1;
}

static CMD_FUNC(cmd_kill)
{
    struct userNode *user;
    if (argc < 3) return 0;
    if (!(user = GetUser(argv[2]))) {
        /* If we get a KILL for a non-local user, it could be a Ghost
         * response to a KILL we sent out earlier.  So we only whine
         * if the target is local. */
        if (!strncmp(argv[2], self->numeric, strlen(self->numeric))) {
            log(MAIN_LOG, LOG_ERROR, "Unable to find kill victim %s\n", argv[2]);
        }
        return 0;
    }

    if (IsLocal(user) && IsService(user)) {
	/* TODO: rate limit this so silly things don't happen. */
	ReintroduceUser(user);
	return 1;
    }

    DelUser(user, NULL, false, argv[2]);
    return 1;
}

static void
join_helper(struct chanNode *cn, void *data)
{
    AddChannelUser(data, cn, 0);
}

static void
join_helper_old_server(char *name, void *data)
{
    struct userNode *un = data;
    struct chanNode *cn;
    cn = AddChannel(name, 0, "", "", 0, un->numeric, NULL);
    if (cn->members.used == 1) {
	struct modeNode *mn = GetUserMode(cn, un);
	if (mn) mn->modes = MODE_CHANOP;
    }
}

static CMD_FUNC(cmd_join)
{
    struct userNode *user;
    if (argc < 3) return 0;
    user = GetUser(argv[0]);
    if (!user) {
        log(MAIN_LOG, LOG_ERROR, "Unable to find user %s joining a channel\n", argv[0]);
        return 0;
    }
    /* Some old servers (like X2) send a JOIN when they put a user into a
     * new channel.  So we have to handle that with join_helper_old_server.
     */
    parse_foreach(argv[2], join_helper, join_helper_old_server, NULL, NULL, user);
    return 1;
}

static void
part_helper(struct chanNode *cn, void *data)
{
    DelChannelUser(data, cn, false, 0);
}

static CMD_FUNC(cmd_part)
{
    struct userNode *user;
    if (argc < 3) return 0;
    user = GetUser(argv[0]);
    parse_foreach(argv[2], part_helper, NULL, NULL, NULL, user);
    return 1;
}

static CMD_FUNC(cmd_kick)
{
    struct userNode *victim;
    struct userNode *kicker;
    struct chanNode *channel;

    if (argc < 4) return 0;
    channel = GetChannel(argv[2]);
    kicker = GetUser(argv[0]);
    victim = GetUser(argv[3]);
    ChannelUserKicked(kicker, victim, channel);
    return 1;
}

static CMD_FUNC(cmd_squit)
{
    const char *server_name, *reason;
    time_t when;
    struct server *server;
    if (argc < 5) return 0;
    server_name = argv[2];
    server = GetServerH(server_name);
    when = strtoul(argv[3], NULL, 10);
    reason = argv[4];
    DelServer(server, 0, reason);
    return 1;
}

static struct {
    chanmsg_func_t func;
    struct userNode *service;
} chanmsg_funcs[256];
static privmsg_func_t *privmsg_funcs;
static unsigned int num_privmsg_funcs = 0;
static privmsg_func_t *notice_funcs;
static unsigned int num_notice_funcs = 0;

struct privmsg_desc {
    struct userNode *user;
    unsigned char *text;
    unsigned int is_notice : 1;
    unsigned int is_qualified : 1;
};

static void
privmsg_chan_helper(struct chanNode *cn, void *data)
{
    struct privmsg_desc *pd = data;

    /* Never send a NOTICE to a channel to one of the services */
    if (!pd->is_notice && chanmsg_funcs[pd->text[0]].func) {
	if (GetUserMode(cn, chanmsg_funcs[pd->text[0]].service)) {
	    chanmsg_funcs[pd->text[0]].func(pd->user, cn, pd->text+1);
	}
    }
}

static void
privmsg_user_helper(struct userNode *un, void *data)
{
    struct privmsg_desc *pd = data;
    unsigned int num = un->num_local;
    if (!pd->is_notice) {
        if ((num < num_privmsg_funcs) && privmsg_funcs[num]) {
            privmsg_funcs[num](pd->user, pd->text, pd->is_qualified);
        }
    } else {
        if ((num < num_notice_funcs) && notice_funcs[num]) {
            notice_funcs[num](pd->user, pd->text, pd->is_qualified);
        }
    }
}

static void
privmsg_invalid(char *name, void *data)
{
    struct privmsg_desc *pd = data;

    irc_401(pd->user, name);
}

static CMD_FUNC(cmd_privmsg)
{
    struct privmsg_desc pd;
    if (argc != 4) return 0;
    pd.user = GetUser(argv[0]);
    if (!pd.user || (IsGagged(pd.user) && !IsOper(pd.user))) return 1;
    pd.is_notice = 0;
    pd.text = argv[3];
    parse_foreach(argv[2], privmsg_chan_helper, NULL, privmsg_user_helper, privmsg_invalid, &pd);
    return 1;
}

static CMD_FUNC(cmd_notice)
{
    struct privmsg_desc pd;
    if (argc != 4 || !argv[0][1] || !argv[0][2]) return 0;
    pd.user = GetUser(argv[0]);
    if (!pd.user || (IsGagged(pd.user) && !IsOper(pd.user))) return 1;
    pd.is_notice = 1;
    pd.text = argv[3];
    parse_foreach(argv[2], privmsg_chan_helper, NULL, privmsg_user_helper, privmsg_invalid, &pd);
    return 1;
}

static CMD_FUNC(cmd_away)
{
    struct userNode *uNode = GetUser(argv[0]);

    if(argc < 3)
    {
	uNode->modes &= ~FLAGS_AWAY;
    }
    else
    {
	uNode->modes |= FLAGS_AWAY;
    }

    return 1;
}

static CMD_FUNC(cmd_gline)
{
    if (argc < 4) return 0;
    if (argv[3][0] == '+') {
        unsigned long duration;
        const char *issuer;
        if (argc < 6) return 0;
        duration = strtoul(argv[4], NULL, 0);
        issuer = (argv[0][0] == ':') ? (argv[0] + 1) : argv[0];
        gline_add(issuer, argv[3]+1, duration, argv[5], 0);
        return 1;
    } else if (argv[3][0] == '-') {
        gline_remove(argv[3]+1, 0);
        return 1;
    } else return 0;
}

static dict_t irc_func_dict;

void
init_parse(void)
{
    irc_func_dict = dict_new();
    dict_insert(irc_func_dict, CMD_BURST, cmd_burst);
    dict_insert(irc_func_dict, TOK_BURST, cmd_burst);
    dict_insert(irc_func_dict, CMD_CREATE, cmd_create);
    dict_insert(irc_func_dict, TOK_CREATE, cmd_create);
    dict_insert(irc_func_dict, CMD_EOB, cmd_eob);
    dict_insert(irc_func_dict, TOK_EOB, cmd_eob);
    dict_insert(irc_func_dict, CMD_EOB_ACK, cmd_eob_ack);
    dict_insert(irc_func_dict, TOK_EOB_ACK, cmd_eob_ack);
    dict_insert(irc_func_dict, CMD_MODE, cmd_mode);
    dict_insert(irc_func_dict, TOK_MODE, cmd_mode);
    dict_insert(irc_func_dict, CMD_NICK, cmd_nick);
    dict_insert(irc_func_dict, TOK_NICK, cmd_nick);
    dict_insert(irc_func_dict, CMD_PASS, cmd_pass);
    dict_insert(irc_func_dict, TOK_PASS, cmd_pass);
    dict_insert(irc_func_dict, CMD_PING, cmd_ping);
    dict_insert(irc_func_dict, TOK_PING, cmd_ping);
    dict_insert(irc_func_dict, CMD_PRIVMSG, cmd_privmsg);
    dict_insert(irc_func_dict, TOK_PRIVMSG, cmd_privmsg);
    dict_insert(irc_func_dict, CMD_PONG, cmd_pong);
    dict_insert(irc_func_dict, TOK_PONG, cmd_pong);
    dict_insert(irc_func_dict, CMD_QUIT, cmd_quit);
    dict_insert(irc_func_dict, TOK_QUIT, cmd_quit);
    dict_insert(irc_func_dict, CMD_SERVER, cmd_server);
    dict_insert(irc_func_dict, TOK_SERVER, cmd_server);
    dict_insert(irc_func_dict, CMD_JOIN, cmd_join);
    dict_insert(irc_func_dict, TOK_JOIN, cmd_join);
    dict_insert(irc_func_dict, CMD_PART, cmd_part);
    dict_insert(irc_func_dict, TOK_PART, cmd_part);
    dict_insert(irc_func_dict, CMD_ERROR, cmd_error);
    dict_insert(irc_func_dict, TOK_ERROR, cmd_error);
    dict_insert(irc_func_dict, CMD_TOPIC, cmd_topic);
    dict_insert(irc_func_dict, TOK_TOPIC, cmd_topic);
    dict_insert(irc_func_dict, CMD_AWAY, cmd_away);
    dict_insert(irc_func_dict, TOK_AWAY, cmd_away);
    dict_insert(irc_func_dict, CMD_SILENCE, cmd_dummy);
    dict_insert(irc_func_dict, TOK_SILENCE, cmd_dummy);
    dict_insert(irc_func_dict, CMD_KICK, cmd_kick);
    dict_insert(irc_func_dict, TOK_KICK, cmd_kick);
    dict_insert(irc_func_dict, CMD_SQUIT, cmd_squit);
    dict_insert(irc_func_dict, TOK_SQUIT, cmd_squit);
    dict_insert(irc_func_dict, CMD_KILL, cmd_kill);
    dict_insert(irc_func_dict, TOK_KILL, cmd_kill);
    dict_insert(irc_func_dict, CMD_NOTICE, cmd_notice);
    dict_insert(irc_func_dict, TOK_NOTICE, cmd_notice);
    dict_insert(irc_func_dict, CMD_STATS, cmd_stats);
    dict_insert(irc_func_dict, TOK_STATS, cmd_stats);
    dict_insert(irc_func_dict, CMD_WHOIS, cmd_whois);
    dict_insert(irc_func_dict, TOK_WHOIS, cmd_whois);
    dict_insert(irc_func_dict, CMD_WALLCHOPS, cmd_notice);
    dict_insert(irc_func_dict, TOK_WALLCHOPS, cmd_notice);
    dict_insert(irc_func_dict, CMD_GLINE, cmd_gline);
    dict_insert(irc_func_dict, TOK_GLINE, cmd_gline);
    /* In P10, DESTRUCT doesn't do anything except be broadcast to servers.
     * Apparently to obliterate channels from any servers that think they
     * exist?
     */
    dict_insert(irc_func_dict, CMD_DESTRUCT, cmd_dummy);
    dict_insert(irc_func_dict, TOK_DESTRUCT, cmd_dummy);
    /* Ignore invites */
    dict_insert(irc_func_dict, CMD_INVITE, cmd_dummy);
    dict_insert(irc_func_dict, TOK_INVITE, cmd_dummy);
    /* DESYNCH is just informational, so ignore it */
    dict_insert(irc_func_dict, CMD_DESYNCH, cmd_dummy);
    dict_insert(irc_func_dict, TOK_DESYNCH, cmd_dummy);
    /* TODO: parse the cmd_dummy lines here */
    dict_insert(irc_func_dict, CMD_VERSION, cmd_dummy);
    dict_insert(irc_func_dict, TOK_VERSION, cmd_dummy);
    dict_insert(irc_func_dict, CMD_ADMIN, cmd_dummy);
    dict_insert(irc_func_dict, TOK_ADMIN, cmd_dummy);
    /* handle topics */
    dict_insert(irc_func_dict, "331", cmd_num_topic);
    dict_insert(irc_func_dict, "332", cmd_num_topic);
    dict_insert(irc_func_dict, "333", cmd_num_topic);
    /* "stats g" responses */
    dict_insert(irc_func_dict, "247", cmd_num_gline);
    dict_insert(irc_func_dict, "219", cmd_dummy); /* "End of /STATS report" */
    /* other numeric responses we might get */
    dict_insert(irc_func_dict, "401", cmd_dummy); /* target left network */
    dict_insert(irc_func_dict, "441", cmd_dummy); /* target isn't on that channel */
    dict_insert(irc_func_dict, "442", cmd_dummy); /* you aren't on that channel */
    dict_insert(irc_func_dict, "443", cmd_dummy); /* is already on channel (after invite?) */

    num_privmsg_funcs = 16;
    privmsg_funcs = malloc(sizeof(privmsg_func_t)*num_privmsg_funcs);
    memset(privmsg_funcs, 0, sizeof(privmsg_func_t)*num_privmsg_funcs);

    num_notice_funcs = 16;
    notice_funcs = malloc(sizeof(privmsg_func_t)*num_notice_funcs);
    memset(notice_funcs, 0, sizeof(privmsg_func_t)*num_notice_funcs);

    userList_init(&dead_users);
};

int
parse_line(unsigned char *line, int recursive)
{
    unsigned char *argv[MAXNUMPARAMS];
    int argc, cmd, res=0;
    cmd_func_t *func;

    argc = split_line(line, true, MAXNUMPARAMS, argv);
    cmd = !burst || !argv[0][1] || !argv[0][2];
    if (argc > cmd) {
        if ((func = dict_find(irc_func_dict, argv[cmd], NULL))) {
            res = func(argc, argv);
        }
    }
    if (!res) {
	unsplit_string(argv, argc);
	log(MAIN_LOG, LOG_ERROR, "PARSE ERROR on line: %s\n", line);
    } else if (!recursive) {
        unsigned int i;
        for (i=0; i<dead_users.used; i++) {
            free(dead_users.list[i]);
        }
        dead_users.used = 0;
    }
    return res;
}

void
reg_privmsg_func(unsigned int numeric, privmsg_func_t handler)
{
    if (numeric > num_privmsg_funcs) {
	int newnum = numeric + 8;
	privmsg_funcs = realloc(privmsg_funcs, newnum*sizeof(privmsg_func_t));
	memset(privmsg_funcs+num_privmsg_funcs, 0, (newnum-num_privmsg_funcs)*sizeof(privmsg_func_t));
	num_privmsg_funcs = newnum;
    }
    if (privmsg_funcs[numeric]) {
	log(DEBUG_LOG, LOG_WARNING, "re-registering new privmsg handler for numeric %d\n", numeric);
    }
    privmsg_funcs[numeric] = handler;
}

void
reg_notice_func(unsigned int numeric, privmsg_func_t handler)
{
    if (numeric > num_notice_funcs) {
	int newnum = numeric + 8;
	notice_funcs = realloc(notice_funcs, newnum*sizeof(privmsg_func_t));
	memset(notice_funcs+num_notice_funcs, 0, (newnum-num_notice_funcs)*sizeof(privmsg_func_t));
	num_notice_funcs = newnum;
    }
    if (notice_funcs[numeric]) {
	log(DEBUG_LOG, LOG_WARNING, "re-registering new notice handler for numeric %d\n", numeric);
    }
    notice_funcs[numeric] = handler;
}

void
reg_chanmsg_func(unsigned char prefix, struct userNode *service, chanmsg_func_t handler)
{
    if (chanmsg_funcs[prefix].func) {
	log(DEBUG_LOG, LOG_WARNING, "re-registering new chanmsg handler for character `%c'\n", prefix);
    }
    chanmsg_funcs[prefix].func = handler;
    chanmsg_funcs[prefix].service = service;
}

void
parse_foreach(char *target_list, foreach_chanfunc cf, foreach_nonchan nc, foreach_userfunc uf, foreach_nonuser nu, void *data)
{
    char *j, old;
    do {
	j = target_list;
	while (*j != 0 && *j != ',') j++;
	old = *j;
        *j = 0;
	if (IsChannelName(target_list)) {
	    struct chanNode *chan = GetChannel(target_list);
            if (chan) {
                if (cf) cf(chan, data);
            } else {
                if (nc) nc(target_list, data);
            }
	} else {
	    struct userNode *user;
            struct privmsg_desc *pd = data;

            pd->is_qualified = 0;
            if (*target_list == '@') {
                user = NULL;
            } else if (strstr(target_list, "@")) {
		struct server *server;

                pd->is_qualified = 1;
		user = GetUserH(strtok(target_list, "@"));
		server = GetServerH(strtok(NULL, "@"));

		if (user && (user->uplink != server)) {
		    /* Don't attempt to index into any arrays
		       using a user's numeric on another server. */
		    user = NULL;
		}
	    } else {
		user = GetUserN(target_list);
	    }

            if (user) {
                if (uf) uf(user, data);
            } else {
                if (nu) nu(target_list, data);
            }
	}
	target_list = j+1;
    } while (old == ',');
}
