#include <iostream>
#include <fstream>
#include <set>
#include <cstdlib>
#include <exception>

extern "C" {
#include "heap.h"
#include "timeq.h"
#include "common.h"

struct timeq_entry {
    timeq_func func;
    void *data;
};

struct heap {
    comparator_f comparator;
    void **data;
    unsigned int data_used, data_alloc;
};

extern heap_t timeq;
};

class timeq_exception {
    const char *_what;
public:
    timeq_exception(const char *what) : _what(what) {}
    virtual const char *what() { return _what; }
    virtual ~timeq_exception() {}
};

/*** TIMEQ INTERFACE DEFINITION ***/

class itimeq {
public:
    itimeq() {}
    virtual void add(time_t when, timeq_func func, void *data) = 0;
    virtual void del(time_t when, timeq_func func, void *data, int mask) = 0;
    virtual time_t next(void) = 0;
    virtual unsigned int size(void) = 0;
    virtual void run(time_t now) = 0;
    virtual void dump(std::ostream &out) = 0;
    virtual void check() = 0;
    virtual ~itimeq() {}
};

/*** C_TIMEQ IMPLEMENTATION ***/

class c_timeq : public itimeq  {
public:
    c_timeq() { timeq_init(); }
    virtual void add(time_t when, timeq_func func, void *data) { timeq_add(when, func, data); }
    virtual void del(time_t when, timeq_func func, void *data, int mask) { timeq_del(when, func, data, mask); }
    virtual time_t next(void) { return timeq_next(); }
    virtual unsigned int size(void) { return timeq_size(); }
    virtual void run(time_t now) { timeq_run(now); }
    virtual void dump(std::ostream &out) {
        unsigned int nn;
        out << "   .. c_timeq (timeq=" << timeq << ") size=" << timeq->data_used
            << " alloc=" << timeq->data_alloc << std::endl;
        for (nn=0; nn<timeq->data_used; nn++) {
            struct timeq_entry *entry;
            entry = (struct timeq_entry*)timeq->data[nn*2+1];
            out << "     .. node " << nn << " when=" << ((time_t)timeq->data[nn*2])
                << " func=" << (void*)entry->func << " data=" << entry->data
                << std::endl;
        }
        out << "   .. end of c_timeq dump" << std::endl;
    }
    virtual void check() {
        unsigned int nn;
        for (nn=0; nn<timeq->data_used; nn++) {
            if ((nn*2+1) >= timeq->data_used) return;
            if ((time_t)timeq->data[(nn*2+1)*2] < (time_t)timeq->data[nn*2]) throw new timeq_exception("left child bad");
            if ((nn*2+2) >= timeq->data_used) return;
            if ((time_t)timeq->data[(nn*2+2)*2] < (time_t)timeq->data[nn*2]) throw new timeq_exception("right child bad");
        }
    }
    virtual ~c_timeq() {}
};

/*** SET_TIMEQ IMPLEMENTATION ***/

class set_timeq : public itimeq  {
    struct entry {
        time_t when;
        timeq_func func;
        void *data;
        bool operator <(const struct entry &other) const { return this->when < other.when; }
        entry(time_t when, timeq_func func, void *data) : when(when), func(func), data(data) {}
    };
    typedef std::multiset<struct entry> set_type;
    set_type timeq2;
public:
    virtual void add(time_t when, timeq_func func, void *data) { timeq2.insert(entry(when, func, data)); }
    virtual void del(time_t when, timeq_func func, void *data, int mask) {
        set_type::iterator it, next;
        for (it=timeq2.begin(); it!=timeq2.end(); it=next) {
            next = it; ++next;
            if (!((mask & TIMEQ_IGNORE_WHEN) || (it->when == when))) continue;
            if (!((mask & TIMEQ_IGNORE_FUNC) || (it->func == func))) continue;
            if (!((mask & TIMEQ_IGNORE_DATA) || (it->data == data))) continue;
            timeq2.erase(it);
        }
    }
    virtual time_t next(void) { return timeq2.begin()->when; }
    virtual unsigned int size(void) { return timeq2.size(); }
    virtual void run(time_t now) {
        set_type::iterator it;
        while (this->size() > 0) {
            it = timeq2.begin();
            if (it->when > now) break;
            timeq2.erase(it);
        }
    }
    virtual void dump(std::ostream &out) {
        set_type::const_iterator it;
        unsigned int nn = 0;
        out << "   .. set_timeq dump; size=" << this->size() << std::endl;
        for (it=timeq2.begin(); it!=timeq2.end(); ++it, ++nn) {
            out << "     .. node " << nn << " when=" << it->when
                << " func=" << (void*)it->func << " data=" << it->data
                << std::endl;
        }
        out << "   .. end of set_timeq dump" << std::endl;
    }
    virtual void check() {}
    virtual ~set_timeq() {}
};

/*** COMPARE_TIMEQ IMPLEMENTATION ***/

class compare_timeq : public itimeq {
public:
    std::fstream logfile;
private:
    class itimeq *tq1, *tq2;
public:
    virtual void check(void) {
        time_t next1, next2;
        unsigned int size1, size2;
        bool error = false;

        next1 = tq1->next();
        next2 = tq2->next();
        if (next1 != next2) {
            logfile << " .. timeqtest_check(): next mismatch: timeq next=" << next1
                    << "; timeq2 next=" << next2 << std::endl;
            error = true;
        }
        size1 = tq1->size();
        size2 = tq2->size();
        if (size1 != size2) {
            logfile << " .. timeqtest_check(): size mismatch: timeq size=" << size1
                    << "; timeq2 size=" << size2 << std::endl;
            error = true;
        }
        try {
            tq1->check();
            tq2->check();
        } catch (...) {
            logfile << " .. timeqtest_check(): check failure" << std::endl;
            error = true;
        }
        if (error) {
            this->dump(logfile);
            throw new timeq_exception("check failure");
        }
    }

public:
    compare_timeq(const char *logfile, itimeq *tq1, itimeq *tq2) : logfile(logfile, std::ios::out), tq1(tq1), tq2(tq2) {
        this->logfile << "Beginning timeq comparison" << std::endl;
    }
    virtual void add(time_t when, timeq_func func, void *data) {
        logfile << "timeq.add(" << when << ", " << (void*)func << ", " << data << ")";
        tq1->add(when, func, data);
        tq2->add(when, func, data);
        logfile << " -> size1=" << tq1->size() << std::endl;
        check();
    }
    virtual void del(time_t when, timeq_func func, void *data, int mask) {
        logfile << "timeq.del(" << when << ", " << (void*)func << ", " << data << ", " << mask << ")";
        tq1->del(when, func, data, mask);
        tq2->del(when, func, data, mask);
        logfile << " -> size1=" << tq1->size() << std::endl;
        check();
    }
    virtual time_t next(void) { check(); return tq1->next(); }
    virtual unsigned int size(void) { check(); return tq1->size(); }
    virtual void run(time_t now) {
        logfile << "timeq.run(" << now << ")" << std::endl;
        tq1->run(now);
        tq2->run(now);
        check();
        logfile << "post-run() size1=" << tq1->size() << "; size2=" << tq2->size() << std::endl;
    }
    virtual void dump(std::ostream &out) {
        out << " .. dump() of tq1:" << std::endl;
        tq1->dump(out);
        out << " .. dump() of tq2:" << std::endl;
        tq2->dump(out);
        out << " .. dump complete." << std::endl;
    }
    virtual ~compare_timeq() { delete tq1; delete tq2; logfile.close(); }
    void dump() { this->dump(logfile); }
};

/*** TESTBENCH DRIVER ***/

void tq_func1(void *data) { (void)data; }
void tq_func2(void *data) { (void)data; }
void tq_func3(void *data) { (void)data; }
void tq_func4(void *data) { (void)data; }
void tq_func5(void *data) { (void)data; }
void tq_func6(void *data) { (void)data; }
void tq_func7(void *data) { (void)data; }
void tq_func8(void *data) { (void)data; }
void tq_func9(void *data) { (void)data; }
void tq_func10(void *data) { (void)data; }

timeq_func tq_funcs[10] = { tq_func1, tq_func2, tq_func3, tq_func4, tq_func5,
                            tq_func6, tq_func7, tq_func8, tq_func9, tq_func10 };

int
main(int argc, char *argv[])
{
    const char *logfile;
    int steps, snap, rnd, nn;
    compare_timeq *tq;
    time_t now;

    now = (argc > 1) ? atoi(argv[1]) : time(NULL); 
    snap = (argc > 2) ? atoi(argv[2]) : -1;
    srand((argc > 3) ? atoi(argv[3]) : now);
    steps = (argc > 4) ? atoi(argv[4]) : 50000;

    logfile = "timeqtest.log";
    tq = new compare_timeq(logfile, new c_timeq(), new set_timeq());

    try {
        for (nn=0; nn<steps; nn++) {
            if (nn == snap) tq->dump();
            rnd = rand() % 100;
            if ((rnd < 60) || (tq->size() == 0)) {
                /* 60% chance: timeq_add() */
                tq->add(now + (rand() % 90) * 3,
                        tq_funcs[rand() % (sizeof(tq_funcs)/sizeof(tq_funcs[0]))],
                        (void*)(rand() % 100));
            } else if (rnd < 90) {
                /* 30% chance: timeq_del() */
                int mask = 0;
                if ((rand() % 16) < 4) mask |= TIMEQ_IGNORE_WHEN;
                if ((rand() % 16) < 4) mask |= TIMEQ_IGNORE_DATA;
                tq->del(now + (rand() % 60),
                        tq_funcs[rand() % (sizeof(tq_funcs)/sizeof(tq_funcs[0]))],
                        (void*)(rand() % 100),
                        mask);
            } else {
                /* 10% chance: timeq_run() */
                now += rand() % 30;
                tq->run(now);
            }
        }
        if (nn == snap) tq->dump();
        std::cout << "Run completed successfully." << std::endl;
        return 0;
    } catch (...) {
        tq->logfile << "Failure at tick " << nn << " with rnd " << rnd << std::endl;
        std::cout << "Exception thrown in tick " << nn << " (rand=" << rnd << ") -- check " << logfile << " for error." << std::endl;
        return 1;
    }
}
